"""
Wrappers for User objects.

.. currentmodule:: curious.dataclasses.user
"""

import datetime
import enum
import typing
from collections import namedtuple
from types import MappingProxyType
from typing import Sequence

from curious.dataclasses import channel as dt_channel, guild as dt_guild, message as dt_message
from curious.dataclasses.appinfo import AppInfo, AuthorizedApp
from curious.dataclasses.bases import Dataclass
from curious.dataclasses.presence import Presence
from curious.exc import CuriousError
from curious.util import AsyncIteratorWrapper, attrdict


class FriendType(enum.IntEnum):
    """
    Represents the type of a friend.
    """
    #: Corresponds to a friend.
    FRIEND = 1

    #: Corresponds to a blocked user.
    BLOCKED = 2

    #: Corresponds to an incoming friend request.
    INCOMING = 3

    #: Corresponds to an outgoing friend request.
    OUTGOING = 4


connection = namedtuple("Connection", "type id name")


class UserProfile(object):
    """
    Represents a profile for a user.
    """

    def __init__(self, user: 'User', kwargs):
        #: The user object associated with this profile.
        self.user = user

        #: A list of connections associated with this profile.
        self.connections = [connection(**c) for c in kwargs.get("connected_accounts", [])]

        #: When was this user premium since?
        self.premium_since = kwargs.get("premium_since", None)

    @property
    def premium(self) -> bool:
        """
        :return: ``True`` if this user has Nitro, ``False`` if not. 
        """
        return self.premium_since is not None or self.user._avatar_hash.startswith("a_")

    def __repr__(self):
        return "<UserProfile user='{}' premium={}>".format(self.user, self.premium)


class User(Dataclass):
    """
    This represents a bare user - i.e, somebody without a guild attached.
    This is used in DMs and similar. All member objects have a reference to their user on ``.user``.

    :ivar id: The ID of this user.
    """

    __slots__ = ("username", "discriminator", "avatar_hash", "verified", "mfa_enabled",
                 "bot", "_bot")

    def __init__(self, client, **kwargs):
        super().__init__(kwargs.get("id"), client)

        #: The username of this user.
        self.username = kwargs.get("username", None)

        #: The discriminator of this user.
        #: Note: This is a string, not an integer.
        self.discriminator = kwargs.get("discriminator", None)

        #: The avatar hash of this user.
        self.avatar_hash = kwargs.get("avatar", None)

        #: If this user is verified or not.
        self.verified = kwargs.get("verified", None)

        #: If this user has MFA enabled or not.
        self.mfa_enabled = kwargs.get("mfa_enabled", None)

        #: If this user is a bot.
        self.bot = kwargs.get("bot", False)

    def _copy(self):
        new_object = object.__new__(self.__class__)
        new_object.id = self.id
        new_object.username = self.username
        new_object.discriminator = self.discriminator
        new_object.avatar_hash = self.avatar_hash
        new_object.verified = self.verified
        new_object.mfa_enabled = self.mfa_enabled
        new_object.bot = self.bot

        new_object._bot = self._bot

        return new_object

    @property
    def avatar_url(self) -> str:
        """
        :return: The avatar URL of this user.
        """
        if not self.avatar_hash:
            return "https://cdn.discordapp.com/embed/avatars/{}.png".format(
                int(self.discriminator) % 5
            )

        # `a_` signifies Nitro and that they have an animated avatar.
        if self.avatar_hash.startswith("a_"):
            suffix = ".gif"  # soon: animated webp
        else:
            suffix = ".webp"

        return "https://cdn.discordapp.com/avatars/{}/{}{}".format(self.id, self.avatar_hash,
                                                                   suffix)

    @property
    def static_avatar_url(self) -> str:
        """
        :return: The avatar URL of this user, but static.
        """
        if not self.avatar_hash:
            return "https://cdn.discordapp.com/embed/avatars/{}.png".format(
                int(self.discriminator) % 5
            )

        return "https://cdn.discordapp.com/avatars/{}/{}.png".format(self.id, self.avatar_hash)

    @property
    def name(self) -> str:
        return self.username

    @property
    def mention(self) -> str:
        """
        :return: A string that mentions this user.
        """
        return "<@{}>".format(self.id)

    @property
    def created_at(self) -> datetime.datetime:
        """
        :return: The time this user was created.
        :rtype: datetime.datetime
        """
        return self.snowflake_timestamp

    def __repr__(self) -> str:
        return "<{} id={} name={} discrim={}>".format(type(self).__name__, self.id, self.name,
                                                      self.discriminator)

    def __str__(self) -> str:
        return "{}#{}".format(self.username, self.discriminator)

    async def open_private_channel(self) -> 'dt_channel.Channel':
        """
        Opens a private channel with a user.

        :return: The newly created private channel.
        """
        if self.discriminator == "0000":
            raise CuriousError("Cannot open a private channel with a webhook")

        # First, try and access the channel from the channel cache.
        original_channel = self._bot.state.find_channel(self.id)
        if original_channel:
            return original_channel

        # Failing that, open a new private channel.
        channel_data = await self._bot.http.create_private_channel(self.id)
        channel = self._bot.state.make_private_channel(channel_data)
        return channel

    async def send(self, content: str = None, *args, **kwargs) -> 'dt_message.Message':
        """
        Sends a message to the user over a private channel.

        :param content: The contet of the message to send.
        :return: A new :class:`~.Message` representing the sent message.
        """
        channel = await self.open_private_channel()
        message = await channel.send(content, *args, **kwargs)

        return message

    async def unban_from(self, guild: 'dt_guild.Guild'):
        """
        Unbans this user from a guild.

        :param guild: The :class:`~.Guild` to unban in.
        """
        return await guild.unban(self)


class UserSettings(attrdict):
    """
    Represents the settings for a user.
    """

    def __init__(self, client, **kwargs):
        self._bot = client

        #: Are emoticons converted automatically?
        #: For example, :) will turn into the emoji literal.
        self.convert_emoticons = kwargs.get("convert_emoticons", True)

        #: Are platform accounts detected?
        self.detect_platform_accounts = kwargs.get("detect_platform_accounts", True)

        #: Is developer mode enabled?
        self.developer_mode = kwargs.get("developer_mode", False)

        #: Is the TTS command enabled?
        self.enable_tts_command = kwargs.get("enable_tts_command", True)

        #: The guild positions for this user.
        self.guild_positions = kwargs.get("guild_positions", [])

        #: The friend source flags for this user.
        self.friend_source_flags = kwargs.get("friend_source_flags", {})

        #: Are inline attachments enabled?
        self.inline_attachment_media = kwargs.get("inline_attachment_media", True)

        #: Are inline embeds enabled?
        self.inline_embed_media = kwargs.get("inline_embed_media", True)

        #: The locale for this user.
        self.locale = kwargs.get("locale", 'en-US')

        #: Is compact mode enabled?
        self.message_display_compact = kwargs.get("message_display_compact", False)

        #: Render embeds?
        self.render_embeds = kwargs.get("render_embeds", True)

        #: Render reactions?
        self.render_reactions = kwargs.get("render_reactions", True)

        #: Restricted guilds (guilds you have DMs disabled for).
        self.restricted_guilds = kwargs.get("restricted_guilds", [])

        #: Show the current game?
        self.show_current_game = kwargs.get("show_current_game", True)

        #: The current theme for this user.
        self.theme = kwargs.get("theme", 'light')

    async def update(self, **kwargs):
        """
        Updates the current user's settings with the keyword args provided.
        
        This will PATCH to Discord, as well.
        """
        # copy a new set of settings and pass it straight to `http.update_user_settings`
        settings = self.copy()
        dict.update(settings, **kwargs)
        # remove this so discord dont get confused
        settings.pop("_bot")

        new_settings = await self._bot.http.update_user_settings(**settings)
        # update ourselves
        dict.update(self, **new_settings)

        return self


class BotUser(User):
    """
    A special type of user that represents ourselves.
    """

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)

        #: The email for this user.
        self.email = kwargs.get("email", None)

        #: Does this user use mobile?
        self.mobile = kwargs.get("mobile", False)

        #: Is this user premium?
        self.premium = kwargs.get("premium", False)

        #: The current :class:`~.UserSettings` for this user.
        self.settings = None  # type: UserSettings

    async def open_private_channel(self):
        raise NotImplementedError("Cannot open a private channel with yourself")

    async def send(self, *args, **kwargs):
        raise NotImplementedError("Cannot send messages to your own user")

    async def block(self, *args, **kwargs):
        raise NotImplementedError("Cannot block or unblock yourself")

    unblock = block

    async def send_friend_request(self):
        raise NotImplementedError("Cannot be friends with yourself")

    remove_friend = send_friend_request

    async def edit(self, *args, **kwargs):
        """
        Edits the bot's current profile.
        """
        return await self._bot.edit_profile(*args, **kwargs)

    async def upload_avatar(self, path: str):
        """
        A higher level interface to editing the bot's avatar.
        """
        return await self._bot.edit_avatar(path)

    @property
    def recent_mentions(self) -> 'typing.AsyncIterator[dt_message.Message]':
        """
        :return: A :class:`~.AsyncIteratorWrapper` that can be used to get all the mentions for \ 
            this user.
        """
        return AsyncIteratorWrapper(self.get_recent_mentions(role_mentions=True,
                                                             everyone_mentions=True,
                                                             limit=100))

    async def get_recent_mentions(self, *,
                                  guild: 'dt_guild.Guild' = None, limit: int=25,
                                  role_mentions: bool=True,
                                  everyone_mentions: bool=True) -> 'Sequence[dt_message.Message]':
        """
        Gets a list of recent mentions for this user. 
        
        :param guild: The guild to limit to. If this is None, it will return ALL recent mentions. 
        :param limit: The maximum amount of mentions to return.
        :param role_mentions: Should role mentions be included?
        :param everyone_mentions: Should everyone mentions be included?
        :return: A list of :class:`~.Message` that represents the recent mentions.
        """
        if self.bot:
            raise CuriousError("Bot accounts cannot use this function")

        if guild is not None:
            guild_id = guild.id
        else:
            guild_id = None

        mentions = await self._bot.http.get_mentions(guild_id=guild_id, limit=limit,
                                                     roles=role_mentions,
                                                     everyone=everyone_mentions)

        messages = []
        for mention in mentions:
            messages.append(self._bot.state.make_message(mention))

        return messages

    @property
    def authorized_apps(self) -> 'typing.AsyncIterator[AuthorizedApp]':
        """
        :return: A :class:`~.AsyncIteratorWrapper` that can be used to get all the authorized apps \ 
            for this user. 
        """
        return AsyncIteratorWrapper(self.get_authorized_apps())

    async def get_authorized_apps(self) -> typing.Sequence[AuthorizedApp]:
        """
        Gets a list of authorized applications for this user.

        :return: A sequence of :class:`~.AuthorizedApp`.
        """
        data = await self._bot.http.get_authorized_apps()
        final = []

        for item in data:
            id = int(item.get("id", 0))
            final.append(AuthorizedApp(id=id, scopes=item.get("scopes", []),
                                       application=AppInfo(self._bot, **item)))

        return final

    @property
    def friends(self) -> typing.Mapping[int, 'RelationshipUser']:
        """
        :return: A mapping of :class:`~.RelationshipUser` that represents the friends for this user.
        """
        if self.bot:
            raise CuriousError("Bots cannot have friends")

        return MappingProxyType(self._bot.state._friends)

    @property
    def blocks(self) -> typing.Mapping[int, 'RelationshipUser']:
        """
        :return: A mapping of :class:`~.RelationshipUser` that represents the blocked users for \ 
            this user.
        """
        if self.bot:
            raise CuriousError("Bots cannot have friends")

        return MappingProxyType(self._bot.state._blocked)


class RelationshipUser(User):
    """
    A user that is either friends or blocked with the current user.
    
    Only useful for user bots.
    """

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)

        #: The presence for this friend.
        self.presence = Presence(status=kwargs.get("status"), game=kwargs.get("game"))

        #: The :class:`.FriendType` of friend this user is.
        self.type_ = None  # type: FriendType

    async def remove_friend(self):
        """
        Removes this user as a friend.
        """
        if self._bot.is_bot:
            raise CuriousError("Bots cannot have friends")

        if self.type_ != FriendType.FRIEND:
            raise CuriousError("This user is not your friend")

        await self._bot.http.remove_relationship(self.id)

    async def unblock(self):
        """
        Unblocks this user.
        """
        if self._bot.is_bot:
            raise CuriousError("Bots cannot have blocks")

        if self.type_ != FriendType.BLOCKED:
            raise CuriousError("This user is not blocked")

        await self._bot.http.remove_relationship(self.id)



"""
Wrappers for custom emojis in guilds.

.. currentmodule:: curious.dataclasses.emoji
"""

import typing

from curious.dataclasses import guild as dt_guild, role as dt_role
from curious.dataclasses.bases import Dataclass


class Emoji(Dataclass):
    """
    Represents a custom emoji uploaded to a guild.
    """
    __slots__ = ("id", "name", "role_ids", "require_colons", "managed", "guild_id")

    def __init__(self, **kwargs):
        super().__init__(int(kwargs.get("id")), kwargs.get("client"))

        #: The name of this emoji.
        self.name = kwargs.get("name", None)

        #: A list of role IDs that this emoji can be used by.
        self.role_ids = kwargs.get("roles", [])

        #: If this emoji requires colons to use.
        self.require_colons: bool = kwargs.get("require_colons", False)

        #: If this emoji is managed or not.
        self.managed: bool = False

        #: The ID of the guild this emoji is associated with.
        self.guild_id: int = None

    def __eq__(self, other):
        if isinstance(other, str):
            return False

        return self.id == other.id

    def __str__(self):
        return "<:{}:{}>".format(self.name, self.id)

    def __repr__(self):
        return "<Emoji guild={} id={}>".format(self.guild, self.id)

    async def edit(self, *,
                   name: str = None,
                   roles: 'typing.List[dt_role.Role]' = None) -> 'Emoji':
        """
        Edits this emoji.

        :param name: The new name of the emoji.
        :param roles: The new list of roles that can use this emoji.
        :return: This emoji.
        """
        if roles is not None:
            roles = [r.id for r in roles]

        await self._bot.http.edit_guild_emoji(guild_id=self.guild_id, emoji_id=self.id,
                                              name=name, roles=roles)
        return self

    async def delete(self):
        """
        Deletes this emoji.
        """
        await self._bot.http.delete_guild_emoji(self.guild_id, emoji_id=self.id)

    @property
    def guild(self) -> 'dt_guild.Guild':
        """
        :return: The :class:`.Guild` this emoji object is associated with.
        """
        return self._bot.guilds.get(self.guild_id)

    @property
    def roles(self) -> 'typing.List[dt_role.Role]':
        """
        :return: A list of :class:`.Role` this emoji can be used by.
        """
        if len(self.role_ids) <= 0:
            return [self.guild.default_role]

        return [self.guild.roles[r_id] for r_id in self.role_ids]

    @property
    def url(self) -> str:
        """
        :return: The URL to this emoji.
        """
        return "https://cdn.discordapp.com/emojis/{}.png".format(self.id)



"""
Commands helpers.

.. currentmodule:: curious.commands

.. autosummary::
    :toctree: Commands

    manager
    context
    decorators
    plugin
    utils

    exc
    converters
"""
from curious.commands.context import Context
from curious.commands.decorators import command, condition
from curious.commands.manager import CommandsManager
from curious.commands.plugin import Plugin



from setuptools import setup

setup(
    name='discord-curious',
    #use_scm_version={
    #    "version_scheme": "guess-next-dev",
    #    "local_scheme": "dirty-tag"
    #},
    version="0.6.0",
    packages=['curious', 'curious.core', 'curious.commands', 'curious.dataclasses',
              'curious.voice', 'curious.ext.loapi', 'curious.ext.paginator'],
    url='https://github.com/SunDwarf/curious',
    license='MIT',
    author='Laura Dickinson',
    author_email='l@veriny.tf',
    description='A curio library for the Discord API',
    setup_requires=[
        "setuptools_scm",
    ],
    zip_safe=True,
    install_requires=[
        "asyncwebsockets>=0.1,<0.2",
        "curio>=0.7.0,<=0.8.0",
        "pylru==1.0.9",
        "oauthlib==2.0.2",
        "pytz",
        "asks>=1.2.2",
        "multidict>=2.1.6"
    ],
    extras_require={
        "voice": ["opuslib==1.1.0",
                  "PyNaCL==1.0.1"],
        "docs": ["guzzle_sphinx_theme", "sphinx", "sphinxcontrib-asyncio",
                 "sphinx-autodoc-typehints"]
    }
)



"""
Class for the commands context.
"""
import inspect
import types
from typing import Any, List, Tuple, Union

from curious.commands.converters import convert_channel, convert_float, convert_int, convert_member
from curious.commands.exc import CommandInvokeError, CommandsError, ConditionsFailedError
from curious.commands.utils import _convert
from curious.core.event import EventContext
from curious.dataclasses.channel import Channel
from curious.dataclasses.guild import Guild
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.user import User


class Context(object):
    """
    A class that represents the context for a command.
    """
    _converters = {
        Channel: convert_channel,
        Member: convert_member,
        # Guild: _convert_guild,
        str: lambda ctx, arg: arg,
        int: convert_int,
        float: convert_float,
    }

    def __init__(self, message: Message, event_context: EventContext):
        """
        :param message: The :class:`.Message` this command was invoked with.
        """
        #: The message for this context.
        self.message: Message = message

        #: The extracted command name for this context.
        self.command_name: str = None

        #: The tokens for this context.
        self.tokens: List[str] = []

        #: The formatted command for this context.
        self.formatted_command: str = None

        #: The plugin for this context.
        self.plugin = None

        #: The manager for this context.
        self.manager = None

        #: The event context for this context.
        self._event_context: EventContext = event_context

        #: The :class:`.Client` for this context.
        self.bot = event_context.bot

    @property
    def guild(self) -> Guild:
        """
        :return: The :class:`.Guild` for this context, or None.
        """
        return self.message.guild

    @property
    def channel(self) -> Channel:
        """
        :return: The :class:`.Channel` for this context.
        """
        return self.message.channel

    @property
    def author(self) -> Union[Member, User]:
        """
        :return: The :class:`.Member` or :class:`.User` for this context.
        """
        return self.message.author

    def match_command(self, func) -> bool:
        """
        Attempts to match a command with this context.
        """
        # don't match subcommands
        if func.cmd_subcommand:
            return False

        # match on name
        if self.command_name == func.cmd_name:
            return True

        # match on alias
        if self.command_name in func.cmd_aliases:
            return True

        # no match
        return False

    def _lookup_converter(self, annotation):
        """
        Looks up a converter for the specified annotation.
        """
        if annotation in self._converters:
            return self._converters[annotation]

        if annotation is inspect.Parameter.empty:
            return lambda ctx, i: i

        # str etc
        if callable(annotation):
            return annotation

        return lambda ctx, i: i

    async def _get_converted_args(self, func) -> Tuple[tuple, dict]:
        """
        Gets the converted args and kwargs for this command, based on the tokens.
        """
        return await _convert(self, self.tokens, inspect.signature(func))

    async def _safety_wrapper(self, coro):
        """
        Runs a command in a safety wrapper.
        """
        try:
            await coro
        except CommandsError as e:
            await self.manager.client.events.fire_event("command_error", self, e, ctx=self)
        except Exception as e:
            try:
                raise CommandInvokeError(self) from e
            except CommandInvokeError as e2:
                await self.manager.client.events.fire_event("command_error", self, e2,
                                                            ctx=self)

    async def can_run(self, cmd) -> Tuple[bool, list]:
        """
        Checks if a command can be ran.

        :return: If it can be ran, and a list of conditions that failed.
        """
        conditions = getattr(cmd, "cmd_conditions", [])
        failed = []
        for condition in conditions:
            try:
                success = condition(self)
                if inspect.isawaitable(success):
                    success = await success
            except CommandsError:
                raise
            except Exception:
                failed.append(condition)
            else:
                if not success:
                    failed.append(success)

        if failed:
            return False, failed

        return True, []

    async def invoke(self, command) -> Any:
        """
        Invokes a command.
        This will convert arguments, pass them in to the command, and run the command.

        :param command: The command function to run.
        """
        # try and do a group lookup
        # how this works:
        # 1) it checks for the current command's subcommands
        # 1a) if empty, it assumes the current matched command is the best we can do
        #     and exits loop
        # 1b) if not empty, it checks all the subcommands for the current command
        # 2) when checking, it checks if the name of the subcommand matches
        # 3a) if it does, set current_command and matched_command, go back to 1
        # 3b) if it doesn't, exit loop so that current_command is the parent command
        matched_command = command
        current_command = command
        # used for subcommands only
        self_ = None
        if hasattr(current_command, "__self__"):
            self_ = current_command.__self__

        self.plugin = self_

        while True:
            if not current_command.cmd_subcommands:
                break

            if not self.tokens:
                break

            token = self.tokens[0]
            for command in current_command.cmd_subcommands:
                if command.cmd_name == token or token in command.cmd_aliases:
                    matched_command = command
                    current_command = command
                    # update tokens so that they're consumed
                    self.tokens = self.tokens[1:]
                    break
            else:
                # we didnt match any subcommand
                # so escape the loop now
                break

        # bind method, if appropriate
        if not hasattr(matched_command, "__self__") and self_ is not None:
            matched_command = types.MethodType(matched_command, self_)

        # check if we can actually run it
        can_run, conditions_failed = await self.can_run(matched_command)
        if not can_run:
            raise ConditionsFailedError(self, conditions_failed)

        # convert all the arguments into the command
        converted_args, converted_kwargs = await self._get_converted_args(matched_command)

        # finally, spawn the new command task
        return await matched_command(self, *converted_args, **converted_kwargs)

    async def try_invoke(self) -> Any:
        """
        Attempts to invoke the command, using the specified manager.

        This will scan all the commands, then invoke as appropriate.
        """
        # temp variable used to invoke if applicable
        to_invoke = None

        for command in self.manager.commands.values():
            if self.match_command(command):
                to_invoke = command
                break

        for plugin in self.manager.plugins.values():
            commands = plugin._get_commands()
            for command in commands:
                if self.match_command(command):
                    to_invoke = command
                    break

        if to_invoke is not None:
            try:
                return await self.invoke(to_invoke)
            except CommandsError as e:
                await self.manager.client.events.fire_event("command_error", e, ctx=self)



"""
Wrappers for Reaction objects.

.. currentmodule:: curious.dataclasses.reaction
"""

import typing

from curious.dataclasses import emoji as dt_emoji


class Reaction(object):
    """
    Represents a reaction.
    """
    def __init__(self, **kwargs):
        #: The :class:`Message` this reaction is for.
        self.message = None

        #: The emoji that represents this reaction.
        self.emoji = None  # type: typing.Union[str, dt_emoji.Emoji]

        #: The number of times this message was reacted to.
        self.count = kwargs.get("count", 1)  # 1 is better than 0

        #: If this user reacted to the message.
        self.me = kwargs.get("me", False)

    def __repr__(self):
        return "<Reaction emoji={} count={}>".format(self.emoji, self.count)

    def __eq__(self, other):
        if not isinstance(other, Reaction):
            return NotImplemented

        if self.message.id != other.message.id:
            return NotImplemented

        return self.emoji == other.emoji

    def __hash__(self):
        # naiive
        return hash(self.message) + hash(self.emoji)



"""
Websocket gateway code.

.. currentmodule:: curious.core.gateway
"""
import collections
import enum
import logging
import sys
import threading
import time
import typing
import zlib

import curio
from asyncwebsockets import Websocket, WebsocketBytesMessage, WebsocketClosed, connect_websocket
from curio.thread import AWAIT, async_thread

try:
    # Prefer ETF data.
    import earl
    _fmt = "etf"


    def _loader(data: bytes):
        return earl.unpack(data, encoding="utf-8", encode_binary_ext=True)


    def _dumper(data: dict):
        return earl.pack(data)

except ImportError:
    try:
        import ujson as json
    except ImportError:
        import json
    finally:
        _fmt = "json"


    def _loader(data: str):
        return json.loads(data)


    def _dumper(data: dict):
        return json.dumps(data)

from curious.dataclasses.presence import Game, Status


# Signalling exceptions.
class ReconnectWebsocket(Exception):
    """
    Signals that the websocket needs to reconnect.
    """


class ChunkGuilds(Exception):
    """
    Signals that we need to begin downloading all guild member chunks.
    """


class GatewayOp(enum.IntEnum):
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE = 3
    VOICE_STATE = 4
    VOICE_PING = 5
    RESUME = 6
    RECONNECT = 7
    REQUEST_MEMBERS = 8
    INVALIDATE_SESSION = 9
    HELLO = 10
    HEARTBEAT_ACK = 11
    GUILD_SYNC = 12


class HeartbeatStats:
    def __init__(self):
        self.heartbeats = 0
        self.heartbeat_acks = 0

        self.last_heartbeat = None
        self.last_ack = None

    @property
    def gw_time(self) -> float:
        """
        :return: The time the most recent heartbeat and heartbeat_ack.
        """
        return self.last_ack - self.last_heartbeat


@async_thread(daemon=True)
def _heartbeat_loop(gw: 'Gateway', heartbeat_interval: float):
    """
    Heartbeat looper that loops and sends heartbeats to the gateway.

    :param gw: The gateway to handle.
    """
    # async threads!
    gw.logger.debug("Sending initial heartbeat.")
    AWAIT(gw.send_heartbeat())
    while True:
        # this is similar to the normal threaded event waiter
        # it will time out after heartbeat_interval seconds.
        try:
            AWAIT(curio.timeout_after(heartbeat_interval, gw._stop_heartbeating.wait()))
        except curio.TaskTimeout:
            pass
        else:
            break

        try:
            AWAIT(gw.send_heartbeat())
        except ReconnectWebsocket:
            break


class Gateway(object):
    """
    Interacts with the Discord Gateway.
    """
    #: The current gateway version to connect to Discord via.
    GATEWAY_VERSION = 6

    def __init__(self, token: str, connection_state, *,
                 large_threshold: int = 250):
        """
        :param token: The bot token to connect with.
        """
        #: The token used to identify with.
        self.token = token

        #: The current websocket connection.
        self.websocket = None  # type: Websocket

        #: The current sequence number.
        self.sequence_num = 0

        #: The connection state used for this connection.
        self.state = connection_state

        #: The shard ID this gateway is representing.
        self.shard_id = 0
        #: The number of shards the client is connected to.
        self.shard_count = 1

        #: The current session ID for this gateway.
        self.session_id = None

        #: The current heartbeat statistic counter for this gateway.
        self.hb_stats = HeartbeatStats()

        #: The current game for this gateway.
        #: Only set when sending a change status packet.
        self.game = None
        #: The current status for this gateway.
        self.status = None

        #: The "large threshold" for this Gateway.
        #: For guilds with members above this number, offline members will not be sent.
        self.large_threshold = min(250, large_threshold)  # bound to 250

        #: The data format for this gateway.
        self.format = _fmt

        self._prev_seq = 0
        self._dispatches_handled = collections.Counter()
        self._enqueued_guilds = []
        self._stop_heartbeating = curio.Event()
        self._logger = None
        self._cached_gateway_url = None  # type: str
        self._open = False

    @property
    def logger(self):
        if self._logger:
            return self._logger

        self._logger = logging.getLogger("curious.gateway:shard-{}".format(self.shard_id))
        return self._logger

    async def connect(self, url: str):
        """
        Connects the client to the websocket.

        This does not do ANY handshake steps.
        :param url: The URL to connect to.
        """
        self.logger.info("Opening connection to {}".format(url))
        try:
            self.websocket = await curio.timeout_after(5, connect_websocket(url))
        except curio.TaskTimeout as e:
            raise ReconnectWebsocket from e
        self.logger.info("Connected to gateway!")
        self._open = True
        return self.websocket

    async def _start_heartbeating(self, heartbeat_interval: float) -> threading.Thread:
        """
        Starts the heartbeat thread.
        :param heartbeat_interval: The number of seconds between each heartbeat.
        """
        if not self._stop_heartbeating.is_set():
            # cancel some poor thread elsewhere
            await self._stop_heartbeating.set()
            del self._stop_heartbeating

        self._stop_heartbeating = curio.Event()

        # dont reference the task - it'll die by itself
        task = await curio.spawn(_heartbeat_loop(self, heartbeat_interval), daemon=True)

        return task

    # Utility methods.
    async def close(self, code: int = 1000, reason: str = "Client disconnected"):
        """
        Closes the connection.

        :param code: The close code.
        :param reason: The close reason.
        """
        if not self.websocket.closed:
            await self.websocket.close(code=code, reason=reason)

        self._open = False
        await self._stop_heartbeating.set()

    async def _send(self, data):
        """
        Sends some data.

        This will actually send the data down the websocket, unlike `send` which only pretends to.
        """
        # Check if heartbeats have drifted.
        if self.hb_stats.heartbeat_acks + 2 < self.hb_stats.heartbeats:
            self.logger.error("Heartbeats have drifted, closing connection!")
            await self.close(code=1000, reason="Failed to receive heartbeat ACKs in time")
            raise ReconnectWebsocket

        try:
            await self.websocket.send_message(data)
        except WebsocketClosed as e:
            await self.close()
            raise

    def _send_dict(self, payload: dict):
        """
        Sends a dict to be packed down the gateway.

        This is a private method - use :meth:`send` instead.

        :param payload: The payload to send.
        """
        data = _dumper(payload)
        return self._send(data)

    async def send(self, data: typing.Any):
        """
        Sends a variable type of data down the gateway.
    """
        if isinstance(data, dict):
            await self._send_dict(data)
        else:
            await self._send(data)

    # Sending events.
    async def send_identify(self):
        """
        Sends an IDENTIFY packet.
        """
        payload = {
            "op": GatewayOp.IDENTIFY,
            "d": {
                "token": self.token,
                "properties": {
                    "$os": sys.platform,
                    "$browser": "curious",
                    "$device": "curious",
                    "$referrer": "",
                    "$referring_domain": ""
                },
                "compress": True,
                "large_threshold": self.large_threshold,
                "v": self.GATEWAY_VERSION,
                "shard": [self.shard_id, self.shard_count]
            }
        }

        await self._send_dict(payload)

    async def send_resume(self):
        """
        Sends a RESUME packet.
        """
        payload = {
            "op": GatewayOp.RESUME,
            "d": {
                "token": self.token,
                "session_id": self.session_id,
                "seq": self.sequence_num
            }
        }

        await self._send_dict(payload)

    async def send_status(self, game: Game, status: Status, *,
                          afk: bool=False):
        """
        Sends a PRESENCE_UPDATE packet.

        :param game: The game object to send.
        :param status: The status object to send.
        :param afk: Is this gateway AFK?
            This should be True for self bots.
        """
        payload = {
            "op": GatewayOp.PRESENCE,
            "d": {
                "game": game.to_dict() if game else None,
                "status": status.value,
                "afk": afk,
                "since": None if not afk else int(time.time() * 1000)
            }
        }

        self.game = game
        self.status = status

        # Update our game() object on all guilds on this shard.
        for guild in self.state.guilds_for_shard(self.shard_id):
            try:
                guild.me.presence.game = game
                guild.me.presence.status = status
            except AttributeError:
                # sent before our member object exists - i.e just after READY happens
                # we can ignore this
                pass

        await self._send_dict(payload)

    async def send_voice_state_update(self, guild_id: int, channel_id: int):
        """
        Sends a voice state update packet.

        :param guild_id: The guild ID to update in.
        :param channel_id: The channel ID to update in.
        """

        payload = {
            "op": GatewayOp.VOICE_STATE,
            "d": {
                "guild_id": str(guild_id),
                "channel_id": str(channel_id),
                "self_mute": False,
                "self_deaf": False,
            }
        }

        await self._send_dict(payload)

    async def send_heartbeat(self):
        """
        Sends a single heartbeat.
        """
        hb = self._get_heartbeat()
        self.logger.debug("Heartbeating with sequence {}".format(hb["d"]))

        # increment the stats
        self.hb_stats.heartbeats += 1
        self.hb_stats.last_heartbeat = time.monotonic()

        await self.state.client.fire_event("gateway_heartbeated", self.hb_stats, gateway=self)

        await self._send_dict(hb)
        return self.hb_stats.heartbeats

    async def send_guild_sync(self, guilds):
        """
        Sends a guild sync packet.

        :param guilds: A list of guild IDs to request syncing for.
        """
        payload = {
            "op": GatewayOp.GUILD_SYNC,
            "d": [str(g.id) for g in guilds]
        }

        await self._send_dict(payload)

    async def request_chunks(self, guilds):
        """
        Requests member chunks from a guild.

        :param guilds: A list of guild IDs to request chunks for.
        """
        payload = {
            "op": GatewayOp.REQUEST_MEMBERS,
            "d": {
                "guild_id": [str(guild.id) for guild in guilds],
                "query": "",
                "limit": 0  # Request ALL!
            }
        }

        await self._send_dict(payload)

    @classmethod
    async def from_token(cls, token: str, state, gateway_url: str,
                         *, shard_id: int = 0, shard_count: int = 1,
                         **kwargs) -> 'Gateway':
        """
        Creates a new gateway connection from a token.

        :param token: The token to pass in.
            This must be a bot user token - user tokens are not supported.
        :param state: The connection state that is attached to this client.
        :param gateway_url: The gateway URL to connect.
        :param shard_id: The shard ID of this bot.
        :param shard_count: The number of shards to start the bot with.

        :return: A new :class:`Gateway` that is connected to the API.
        """
        obb = cls(token, state, **kwargs)
        obb.shard_id = shard_id
        obb.shard_count = shard_count

        gateway_url += "/?v={}&encoding={}".format(cls.GATEWAY_VERSION, _fmt)
        obb._cached_gateway_url = gateway_url

        await obb.connect(gateway_url)

        # send IDENTIFY
        obb.logger.info("Sending IDENTIFY...")
        await obb.send_identify()

        return obb

    async def reconnect(self, *, resume: bool = False):
        """
        Reconnects the bot to the gateway.

        :param resume: Should a RESUME be attempted?
        """
        self.logger.info("Reconnecting to the gateway")

        # reset our heartbeat count
        self.hb_stats.heartbeats = 0
        self.hb_stats.heartbeat_acks = 0

        if not self.websocket.closed:
            await self.websocket.close(code=1001, reason="Forcing a reconnect")

        self._open = False

        await self.connect(self._cached_gateway_url)

        if resume:
            # Send the RESUME packet, instead of the IDENTIFY packet.
            self.logger.info("Sending RESUME...")
            await self.send_resume()
        else:
            self.logger.info("Sending IDENTIFY...")
            self.sequence_num = 0
            await self.send_identify()

        return self

    async def _get_chunks(self):
        """
        Called to start chunking all guild members.
        """
        for guild in self._enqueued_guilds:
            guild.start_chunking()
            self.logger.info("Requesting {} member chunk(s) from guild {}."
                             .format(guild._chunks_left, guild.name))

        await self.request_chunks(self._enqueued_guilds)
        self._enqueued_guilds.clear()

    def _get_heartbeat(self):
        return {
            "op": GatewayOp.HEARTBEAT,
            "d": self.sequence_num
        }

    async def next_event(self):
        """
        Gets the next event, in decoded form.
        """
        if not self._open:
            raise WebsocketClosed(1006, reason="Connection lost")

        try:
            event = await self.websocket.next_message()
        except WebsocketClosed:
            # Close ourselves.
            await self.close()
            raise

        await self.state.client.fire_event("gateway_message_received", event, gateway=self)

        if isinstance(event, WebsocketBytesMessage) and _fmt == "json":
            # decompress the message
            data = zlib.decompress(event.data, 15, 10490000)
            data = data.decode("utf-8")
        else:
            data = event.data

        if data is None:
            return

        event_data = _loader(data)
        # self.logger.debug("Got event {}".format(event_data))
        await self.state.client.fire_event("gateway_event_received", event_data, gateway=self)

        op = event_data.get('op')
        data = event_data.get('d')
        seq = event_data.get('s')

        if seq is not None:
            # next sequence
            self.sequence_num = seq

        # Switch based on op.
        if op == GatewayOp.HELLO:
            # Start heartbeating, with the specified heartbeat duration.
            await self.state.client.fire_event("gateway_hello", data['_trace'], gateway=self)
            heartbeat_interval = data.get("heartbeat_interval", 45000) / 1000.0
            self.logger.debug("Heartbeating every {} seconds.".format(heartbeat_interval))
            await self._start_heartbeating(heartbeat_interval)
            self.logger.info("Connected to Discord servers {}".format(",".join(data["_trace"])))

        elif op == GatewayOp.HEARTBEAT_ACK:
            await self.state.client.fire_event("gateway_heartbeat_ack", gateway=self)
            self.hb_stats.heartbeat_acks += 1
            self.hb_stats.last_ack = time.monotonic()

        elif op == GatewayOp.HEARTBEAT:
            # Send a heartbeat back.
            await self.state.client.fire_event("gateway_heartbeat_received", gateway=self)
            await self.send_heartbeat()

        elif op == GatewayOp.INVALIDATE_SESSION:
            # Clean up our session.
            should_resume = data
            await self.state.client.fire_event("gateway_invalidate_session", should_resume,
                                               gateway=self)
            if should_resume is True:
                self.logger.debug("Sending RESUME again")
                await self.send_resume()
            else:
                self.logger.warning("Received INVALIDATE_SESSION with d False, re-identifying.")
                self.sequence_num = 0
                self.state._reset(self.shard_id)
                await self.send_identify()

        elif op == GatewayOp.RECONNECT:
            # Try and reconnect to the gateway.
            await self.state.client.fire_event("gateway_reconnect_received", gateway=self)
            self.close()
            raise ReconnectWebsocket()

        elif op == GatewayOp.DISPATCH:
            # Handle the dispatch.
            event = event_data.get("t")
            handler = getattr(self.state, "handle_{}".format(event.lower()), None)

            if handler:
                self.logger.debug("Parsing event {}.".format(event))
                self._dispatches_handled[event] += 1
                await self.state.client.fire_event("gateway_dispatch_received", data, gateway=self)
                # Invoke the handler, which will parse the data and update the cache internally.
                # Yes, handlers are async.
                # This is because curio requires `spawn()` to be async.
                try:
                    await handler(self, data)
                except ChunkGuilds as e:
                    # We need to download all member chunks from this guild.
                    await self._get_chunks()
                except Exception:
                    self.logger.exception("Error decoding event {} with data "
                                          "{}".format(event, data))
                    await self.close(code=1006, reason="Client error")
                    raise
            else:
                self.logger.warning("Unhandled event: {}".format(event))

        else:
            try:
                self.logger.warning("Unhandled opcode: {} ({})".format(op, GatewayOp(op)))
            except ValueError:
                self.logger.warning("Unknown opcode: {}".format(op))
